#!/bin/bash
#
# Copyright (c) 2012-2014 Fixpoint Inc. All rights reserved.
# ---
# Kompira簡易インストーラ：
#   - 1台構成版
#   - Apache/PostgreSQLを使用
#
# 動作要件：
#   - x86_64
#   - CentOS 7.x/6.x系 (7.x 推奨)
#   - Python 2.7系
#
shopt -s extglob
SETUP_TYPE="Install the Kompira"
INSTALL_LOG="install.$$.log"
THIS_DIR=$(dirname $(readlink -f $0))

. $THIS_DIR/scripts/setup_utils.sh

KOMPIRA_LOGIN=/bin/bash
KOMPIRA_PKGNAME_DEFAULT=Kompira
KOMPIRA_PKGNAME_JOBMNGR=Kompira-jobmngr
KOMPIRA_PKGNAME_SENDEVT=Kompira-sendevt
KOMPIRA_WHLFILE_DEFAULT="$THIS_DIR/dist/Kompira-$VERSION-py2-none-any.whl"
KOMPIRA_WHLFILE_JOBMNGR="$THIS_DIR/dist/Kompira_jobmngr-$VERSION-py2-none-any.whl"
KOMPIRA_WHLFILE_SENDEVT="$THIS_DIR/dist/Kompira_sendevt-$VERSION-py2-none-any.whl"
KOMPIRA_TEST_URL="https://localhost/.login"

#
# HTTP(S) アクセス可能かチェックする URL
#
CHECK_ACCESSIBILITY="https://bootstrap.pypa.io https://yum.postgresql.org https://dl.fedoraproject.org"

#
# SCL レポジトリ情報
#
SCL_CENT_REPOPKG="centos-release-scl-rh"
SCL_RHEL_REPONAME="rhel-server-rhscl-${RHEL_VERSION}-rpms"
SCL_RHUI_REPONAME="rhui-rhel-server-rhscl-${RHEL_VERSION}-rhui-rpms"
SCL_PYTHON27_ROOT=/opt/rh/python27/root

#
# RHEL レポジトリ情報
#
RHEL_OPTION_REPONAME="rhel-${RHEL_VERSION}-server-optional-rpms"
RHUI_OPTION_REPONAME="rhui-rhel-${RHEL_VERSION}-server-rhui-optional-rpms"

#
# postgres レポジトリ／パッケージ情報
#
PG_YUMREPO="https://yum.postgresql.org/${PG_VERSION}/redhat/rhel-${RHEL_VERSION}-${REQUIRES_ARCH}/${PG_REPOPKG}-${PG_VERSION}-.*.noarch.rpm"

#
# epel レポジトリ情報
#
EPEL_REPONAME="epel"
EPEL_REPOPKG="epel-release"
EPEL_YUMREPO="https://dl.fedoraproject.org/pub/epel/epel-release-latest-${RHEL_VERSION}.noarch.rpm"

#
# localhost チェック用ワンライナー
#
PYTHON_CHECK_LOCALHOST='from socket import *; [socket(*res[:3]) and exit(0) for res in getaddrinfo("localhost", 0, 0, SOCK_STREAM)] or exit(1)'

#
# pip のインストールスクリプト
#
: ${GETPIP_PY:="https://bootstrap.pypa.io/get-pip.py"}

#
# firewalld/iptables の許可ポート
#
FIREWALLD_ACCEPT_PORTS="http https 5672/tcp"
IPTABLES_ACCEPT_PORTS="80/tcp 443/tcp 5672/tcp"

#
# SELinux のモード
#
SELINUX_MODE=permissive

#
# インストール自体に必要なパッケージ
#
PACKAGES_FOR_INSTALL="yum-utils curl nss"

#
# kompira_sendevt に必要なパッケージ
#
PACKAGES_FOR_SENDEVT=""
PACKAGES_FOR_SENDEVT_ONLINE=""

#
# kompira_jobmngr に必要なパッケージ
#
# gcc: pycrypto
# openssl-devel:
# krb5-devel は openssl-devel に依存してインストールされる
PACKAGES_FOR_JOBMNGR="krb5-workstation openssl libffi"
PACKAGES_FOR_JOBMNGR_ONLINE="gcc make krb5-devel openssl-devel libffi-devel"
if ! $SYSTEMD; then
    PACKAGES_FOR_JOBMNGR="$PACKAGES_FOR_JOBMNGR upstart"
fi

#
# kompira サーバに必要なパッケージ
#
PACKAGES_FOR_KOMPIRA="httpd mod_ssl mercurial git sudo iproute jq"
PACKAGES_FOR_KOMPIRA_ONLINE="httpd-devel"
PACKAGES_FOR_KOMPIRA_RABBITMQ="rabbitmq-server"
PACKAGES_FOR_KOMPIRA_POSTGRESQL="${PG_PREFIX}-server ${PG_PREFIX}-contrib"
#PACKAGES_FOR_KOMPIRA_POSTGRESQL_ONLINE="${PG_PREFIX}-devel"

#
# extra パッケージ構築に必要なパッケージ
#
PACKAGES_FOR_EXTRA="createrepo $PACKAGES_FOR_JOBMNGR_ONLINE $PACKAGES_FOR_KOMPIRA_ONLINE"
PACKAGES_FOR_OFFLINE=""

#
# ロケールの設定
#
: ${CONFIG_LOCALE_LANG:=false}
: ${CONFIG_LOCALE_TIMEZONE:=false}
: ${LOCALE_LANG:=""}
: ${LOCALE_TIMEZONE:="Asia/Tokyo"}
: ${LOCALE_CLOCK_UTC:="false"}

#
# LOCALE_LANG のデフォルトは postgres の locale 設定があればそれを引き継ぐ
# なければ現状の LANG を引き継ぐ
#
if [ -z "$LOCALE_LANG" ]; then
    PGSQL_LANG=$(psql -U $PGSQL_USER -d $PGSQL_DB -At -c "show lc_collate;" 2>/dev/null)
    if [ -n "$PGSQL_LANG" ]; then
        LOCALE_LANG=$PGSQL_LANG
    else
        LOCALE_LANG=$LANG
    fi
fi

#
# インストール時の proxy の設定
#
: ${TEMP_PROXY_URL:="$PROXY_URL"}
: ${TEMP_NO_PROXY:="$NO_PROXY"}

HTTPS_MODE=true
FORCE_MODE=false
BACKUP_MODE=true
BACKUP_PROCESS=false
INITDATA_MODE=false
INITFILE_MODE=false
MIGRATE_FAKE=false
EXTRA_MODE=false
JOBMNGR_MODE=false
SENDEVT_MODE=false
USER_MODE=false
KOMPIRA_SERVER=localhost
REQUIRES_ONLY=false
DRY_RUN=false
SHOW_OPTIONS="$SHOW_OPTIONS LOCALE_LANG PATH HTTPS_MODE FORCE_MODE BACKUP_MODE BACKUP_PROCESS INITDATA_MODE INITFILE_MODE OFFLINE_MODE JOBMNGR_MODE SENDEVT_MODE USER_MODE PROXY_URL NO_PROXY KOMPIRA_SERVER DRY_RUN"

check_localhost()
{
    $PYTHON -c "$PYTHON_CHECK_LOCALHOST" > /dev/null 2>&1
}

check_environ()
{
    echo_title "Check current environment."
    if [ -n "$TEMP_PROXY_URL" ]; then
        # インストール時のプロキシ環境変数を設定する
        echo_debug "settings http_proxy for install: $TEMP_PROXY_URL"
        echo_debug "settings no_proxy for install: $TEMP_NO_PROXY"
        export http_proxy="$TEMP_PROXY_URL"
        export https_proxy="$TEMP_PROXY_URL"
        export HTTP_PROXY="$TEMP_PROXY_URL"
        export HTTPS_PROXY="$TEMP_PROXY_URL"
        export no_proxy="$TEMP_NO_PROXY"
    fi

    check_arch
    check_system
    check_subscription $OFFLINE_MODE
    if $EXTRA_MODE; then return 0; fi

    if ! $OFFLINE_MODE; then
        # ネットワークアクセスできるか確認する
        echo_info "Check network accessibility"
        local index=0
        for url in $CHECK_ACCESSIBILITY; do
            let index=$index+1
            verbose_run curl -x "$TEMP_PROXY_URL" --connect-timeout 10 -kvfI "$url" > /dev/null 2> $TMPDIR/test-accessibility$index.out
            if [ $? != 0 ]; then
                cat $TMPDIR/test-accessibility$index.out
                echo_error "[ERROR] Can not connect to $url"
                echo_error "Please make sure this server can connect to the internet. Or check the proxy settings."
                abort_setup
            fi
            echo_info "OK: $url"
        done
    fi

    if $OFFLINE_MODE; then
        # オフラインインストールに必要な extra パッケージが展開されているかチェック
        local dir
        if [ ! -f $KOMPIRA_EXTRA_REPO ]; then
            abort_setup "$KOMPIRA_EXTRA_REPO not found!"
        fi
        for dir in $KOMPIRA_EXTRA_PIP $KOMPIRA_EXTRA_RPM $KOMPIRA_EXTRA_WHL; do
            if [ ! -d $dir ]; then
                abort_setup "$dir not found!"
            fi
        done
    fi
    if ! $JOBMNGR_MODE && ! $SENDEVT_MODE; then
        # 旧バージョンの PostgreSQL チェック
        if rpm -q postgresql > /dev/null 2>&1; then
            abort_setup "Old postgresql is installed." \
                        "Please uninstall postgreql and re-execute."
        fi
        # localhost に対して正しいアドレス情報を返すことをチェック
        if ! check_localhost; then
            abort_setup "The address of localhost is invalid." \
                        "Please make sure localhost in set in /etc/hosts"
        fi
    fi
}

check_cluster()
{
    if $EXTRA_MODE; then return 0; fi
    echo_title "Check cluster status."
    get_cluster_status
    if $CLUSTER_RUNNING; then
        # 冗長構成が実行中の場合
        #   マスターノードではインストールを禁止する
        #   スレーブノードでは --force オプションをチェックする
        if $CLUSTER_MASTER; then
            abort_setup "Pacemaker is runnning (MASTER), so cannot install." \
                        "To install, demote to slave or stop the pacemaker."
        elif ! $FORCE_MODE; then
            abort_setup "Pacemaker is runnning (SALVE), so cannot install." \
                        "To install, use --force option."
        fi
    fi
}

check_python()
{
    install_python27

    echo_title "Check python version."
    local result=$($PYTHON -V 2>&1 >/dev/null)
    case $result in
        $REQUIRES_PYTHON27)
            echo_info "Confirmed Python version: $result"
            echo_info "PYTHON=$PYTHON"
            ;;
        *)
            abort_setup "$REQUIRES_PYTHON27 is required"
            ;;
    esac
}

check_kompira_package()
{
    if $EXTRA_MODE; then return 0; fi
    local -x PIP=$(which pip 2> /dev/null)
    if [ -z "$PIP" ]; then
        return 0
    fi
    OLD_KOMPIRA_VERSION="$(get_kompira_version)"
    echo_info "VERSION=$OLD_KOMPIRA_VERSION [pip=$PIP]"
    case $OLD_KOMPIRA_VERSION in
        "")
            : # echo "*** Kompira はインストールされていません"
            ;;
        ${BRANCH}*)
            echo_info "A compatible version is installed."
            ;;
        *)
            if ! $FORCE_MODE; then
                abort_setup "An incompatible version is installed, so cannot update." \
                            "To install, use --force option."
            elif ! $INITDATA_MODE; then
                echo_warn "An incompatible version is installed, so force to INITDATA_MODE"
                INITDATA_MODE=true
            fi
            ;;
    esac
}

check_update_kompira()
{
    # INITDATA モードでは Kompira の互換性チェックは不要
    if $INITDATA_MODE; then return 0; fi

    echo_title "Check version of Kompira installed."

    # システム環境にインストールされた Kompira (obsoleted) をチェック
    check_kompira_package

    # virtualenv 環境にインストールされている Kompira をチェック
    PATH=$KOMPIRA_BIN/:$PATH \
    check_kompira_package
}

check_update_pgsql()
{
    echo_title "Check PostgreSQL verison and Kompira database existence."
    #
    # 動作中の postgresql を検出する
    # [AMZN] /usr/lib64/pgsql9x/bin/postmaster
    # [CENT] /usr/pgsql-9.x/bin/postmaster
    #
    OLD_PG_VERSION=$(readlink "/proc/$(pgrep -o postmaster)/exe" | sed -re 's|.*/pgsql-?([0-9.]+)/.*|\1|')
    if [ -n "$OLD_PG_VERSION" ]; then
        if [ "$SYSTEM" == "AMZN" ]; then
            OLD_PG_SERVICE="postgresql${OLD_PG_VERSION/./}"
        else
            OLD_PG_SERVICE="postgresql-${OLD_PG_VERSION}"
        fi
        if [ "$OLD_PG_SERVICE" != "$PG_SERVICE" ]; then
            if $CLUSTER_RUNNING; then
                abort_setup "Detected postgresql ${OLD_PG_VERSION} and Clustering is running, so cannot update." \
                            "To install, stop the pacemaker."
            fi
            echo_warn "Detected postgresql ${OLD_PG_VERSION}, so force to INITDATA_MODE"
            INITDATA_MODE=true
            return 0
        fi
    fi
    #
    # INITDATA モードでは Kompira データベースが存在する場合、消してよいか確認する
    #
    if ! $INITDATA_MODE; then
        return 0
    fi
    local PSQL=$(which psql 2> /dev/null)
    if [ -z "$PSQL" ]; then
        return 0
    fi
    local exist=false
    if $PSQL -U $PGSQL_USER -d $KOMPIRA_PG_DATABASE -c \\q; then
        echo_warn "Kompira Database '$KOMPIRA_PG_DATABASE' is exist already"
        exist=true
    fi
    if $PSQL -U $KOMPIRA_PG_USER -d $PGSQL_DB -c \\q; then
        echo_warn "Kompira DB-user '$KOMPIRA_PG_USER' is exist already"
        exist=true
    fi
    if $FORCE_MODE; then
        # 強制モードでは既存ユーザ／データベースがあっても継続する
        return 0
    fi
    if $exist; then
        echo
        inquire "DELETE AND CONTINUE INSTALLATION? (yes/No)" "n"
        echo $answer
        echo
        if [ $answer != "y" ]; then
            abort_setup "Abort installation."
        fi
    fi
}

check_update()
{
    if $JOBMNGR_MODE || $SENDEVT_MODE; then return 0; fi
    if $EXTRA_MODE; then return 0; fi

    # 実行中の pgsql のバージョンおよび既存データベースをチェックする
    check_update_pgsql
    # インストールされている Kompira の互換性チェックする
    check_update_kompira
}

yum_clean()
{
    if $SKIP_CLEAN_YUM_CACHE; then return 0; fi
    echo_title "Clean yum cache."

    # 重複パッケージを削除する
    if [ $RHEL_VERSION == 6 ]; then
        verbose_run package-cleanup -y --cleandupes
    elif [ $RHEL_VERSION == 7 ]; then
        verbose_run package-cleanup -y --cleandupes --removenewestdupes
    fi

    # yum check-upate がキャッシュを参照しないようにクリアしておく
    verbose_run yum clean all > /dev/null

    if ! $OFFLINE_MODE; then
        echo_info "Check yum update."
        verbose_run yum -q check-update > /dev/null
        if [ "$?" -eq 1 ]; then
            echo_error "[ERROR] Can not connect to yum repository."
            echo_error "Please make sure this server can connect to the internet. Or check the proxy settings."
            abort_setup
        fi
    fi
}

configure_locale_lang()
{
    if ! $CONFIG_LOCALE_LANG; then return 0; fi
    if $EXTRA_MODE; then return 0; fi

    echo_title "Check locale language settings"
    if $SYSTEMD; then
        if LANG= localectl status | grep -q -F "Locale: LANG=$LOCALE_LANG"; then
            echo_info "Confirmed language: $LOCALE_LANG"
        else
            echo_info "Change language setting ($LANG -> $LOCALE_LANG)"
            verbose_run localectl set-locale LANG=$LOCALE_LANG
            LANG=$LOCALE_LANG
        fi
    else
        if [ "$LANG" == "$LOCALE_LANG" ]; then
            echo_info "Confirmed LANG: $LANG"
        else
            echo_info "Change LANG setting ($LANG -> $LOCALE_LANG)"
            LANG=$LOCALE_LANG
        fi
        if grep -F "LANG=\"$LOCALE_LANG\"" /etc/sysconfig/i18n > /dev/null; then
            echo_info "Confirmed /etc/sysconfig/i18n: $LOCALE_LANG"
        else
            echo_info "Change /etc/sysconfig/i18n settings ($LOCALE_LANG)"
            echo "LANG=\"$LOCALE_LANG\"" > $TMPDIR/i18n
            diff_cp $TMPDIR/i18n /etc/sysconfig/i18n
        fi
    fi
}

configure_locale_timezone()
{
    if ! $CONFIG_LOCALE_TIMEZONE; then return 0; fi
    if $EXTRA_MODE; then return 0; fi

    echo_title "Check locale timezone settings"
    if $SYSTEMD; then
        if LANG= timedatectl status | grep -q -F "Time zone: $LOCALE_TIMEZONE"; then
            echo_info "Confirmed timezone: $LOCALE_TIMEZONE"
        else
            echo_info "Change timezone settings ($LOCALE_TIMEZONE)"
            verbose_run timedatectl set-timezone $LOCALE_TIMEZONE
        fi
    else
        if diff -q /usr/share/zoneinfo/$LOCALE_TIMEZONE /etc/localtime > /dev/null; then
            echo_info "Confirmed /etc/localtime: $LOCALE_TIMEZONE"
        else
            echo_info "Change /etc/localtime settings ($LOCALE_TIMEZONE)"
            verbose_run cp /usr/share/zoneinfo/$LOCALE_TIMEZONE /etc/localtime
            service_restart crond
        fi
        if grep -F "ZONE=\"$LOCALE_TIMEZONE\"" /etc/sysconfig/clock > /dev/null; then
            echo_info "Confirmed /etc/sysconfig/clock: $LOCALE_TIMEZONE"
        else
            echo_info "Change /etc/sysconfig/clock settings ($LOCALE_TIMEZONE)"
            echo -e "ZONE=\"$LOCALE_TIMEZONE\"\nUTC=$LOCALE_CLOCK_UTC" > $TMPDIR/clock
            diff_cp $TMPDIR/clock /etc/sysconfig/clock
        fi
    fi
}

create_user()
{
    echo_title "Create Kompira account: $KOMPIRA_USER"
    if ! id $KOMPIRA_USER > /dev/null 2>&1; then
        verbose_run useradd -s $KOMPIRA_LOGIN $KOMPIRA_USER
    else
        verbose_run usermod -s $KOMPIRA_LOGIN $KOMPIRA_USER
    fi
    #
    # .bash_profile に virtualenv を activate する設定を追加
    # MEMO: チルダ展開の前にパラメータ展開するため eval で展開
    #
    local BASH_PROFILE=$(eval "echo ~$KOMPIRA_USER/.bash_profile")
    cp $BASH_PROFILE $TMPDIR/bash_profile

    if grep -E -i -q "^export http?_proxy=" $TMPDIR/bash_profile; then
        verbose_run \
            sed -i -r \
            -e "s|^(export https?_proxy)=.*|\1=\"$PROXY_URL\"|i" \
            -e "s|^(export no_proxy)=.*|\1=\"$NO_PROXY\"|" \
            $TMPDIR/bash_profile
    else
        cat >> $TMPDIR/bash_profile <<EOF

# proxy settings
export http_proxy="$PROXY_URL"
export https_proxy="$PROXY_URL"
export HTTP_PROXY="$PROXY_URL"
export HTTPS_PROXY="$PROXY_URL"
export no_proxy="$NO_PROXY"
EOF
    fi
    if ! grep -q -F "source $KOMPIRA_BIN/activate" $TMPDIR/bash_profile; then
        cat >> $TMPDIR/bash_profile <<EOF

# activate virtualenv
if [ -f $KOMPIRA_BIN/activate ]; then
    source $KOMPIRA_BIN/activate
fi
EOF
    fi
    diff_install -o $KOMPIRA_USER -g $KOMPIRA_GROUP $TMPDIR/bash_profile $BASH_PROFILE
}

USER_SUDO()
{
    if $USER_MODE; then
        verbose_run sudo -u $KOMPIRA_USER "$@"
    else
        # bash -c には引数をまとめて渡す必要があるので "$*" で展開する
        verbose_run bash -c "$*"
    fi
}

USER_INSTALL()
{
    if $USER_MODE; then
        verbose_run $INSTALL -o $KOMPIRA_USER -g $KOMPIRA_GROUP "$@"
    else
        verbose_run $INSTALL "$@"
    fi
}

create_home()
{
    echo_title "Create Kompira directory: $KOMPIRA_HOME"
    USER_INSTALL -m 755 -d $KOMPIRA_HOME
}

create_userbase()
{
    echo_title "Create userbase directory, and install pip/virtualenv"
    #
    # pip/virtualenv は常に ~kompira/.local にインストールするため USER_SUDO は使わない
    #
    local KOMPIRA_USERBASE=$(sudo -u $KOMPIRA_USER $PYTHON -c 'import site; print(site.USER_BASE)')
    local install_pip=true
    local pip_options=""
    if $OFFLINE_MODE; then
        pip_options="$pip_options --no-index --find-links=$KOMPIRA_EXTRA_WHL"
    fi
    if [ -n "$TEMP_PROXY_URL" ]; then
        pip_options="$pip_options --proxy=$TEMP_PROXY_URL"
    fi
    #
    # $KOMPIRA_USERBASE/bin/pip が（あれば） python 2.6 用かチェックする
    #
    if [ -x $KOMPIRA_USERBASE/bin/pip ]; then
        local pip_version=$(sudo -u $KOMPIRA_USER $KOMPIRA_USERBASE/bin/pip --version)
        if echo "$pip_version" | grep -q "python.*2.6"; then
            echo_debug "pip for python 2.6 detected: $pip_version"
        else
            install_pip=false
        fi
    fi
    #
    # get-pip.py で $KOMPIRA_USERBASE/bin に pip をインストールする
    # すでにシステム領域に pip がインストールされていても無視するように -I オプションをつける
    #
    if $install_pip; then
        echo_info "Install pip to $KOMPIRA_USERBASE/bin"
        local user_tmpdir=$(sudo -u $KOMPIRA_USER mktemp -d)
        local errmsg="Failed to download get-pip.py"
        if $OFFLINE_MODE; then
            cp $KOMPIRA_EXTRA_PIP/get-pip.py $user_tmpdir/get-pip.py
        else
            verbose_run curl -ksL -o $user_tmpdir/get-pip.py $GETPIP_PY
        fi
        local rc=$?
        if [ $rc -eq 0 ]; then
            verbose_run sudo -u $KOMPIRA_USER $PYTHON $user_tmpdir/get-pip.py $pip_options -I -U --user --no-setuptools --no-wheel
            rc=$?
            errmsg="Failed to install pip"
        fi
        verbose_run rm -rf $user_tmpdir
        exit_if_failed $rc "$errmsg"
    fi
    #
    # ~kompira/.local の pip 用 proxy 設定
    #
    echo_info "Configure ~$KOMPIRA_USER/.config/pip/pip.conf"
    local USER_PIP_CONFIGDIR=$(eval "echo ~$KOMPIRA_USER/.config/pip")
    verbose_run sudo -u $KOMPIRA_USER $INSTALL -d $USER_PIP_CONFIGDIR
    echo -e "[global]\nproxy=$PROXY_URL" > $TMPDIR/userenv_pip.conf
    diff_install -o $KOMPIRA_USER -g $KOMPIRA_GROUP $TMPDIR/userenv_pip.conf $USER_PIP_CONFIGDIR/pip.conf
    #
    # ~kompira/.local の pip をアップデート
    #
    if ! $install_pip; then
        echo_info "Update pip to $KOMPIRA_USERBASE/bin"
        verbose_run sudo -u $KOMPIRA_USER $KOMPIRA_USERBASE/bin/pip install $pip_options -I -U --user pip
    fi
    echo_info "Install virtualenv to $KOMPIRA_USERBASE/bin"
    verbose_run sudo -u $KOMPIRA_USER $KOMPIRA_USERBASE/bin/pip install $pip_options -I -U --user --no-warn-script-location virtualenv
    exit_if_failed "$?" "Failed to install virtualenv"
}

create_virtualenv()
{
    echo_title "Crete virtualenv: $KOMPIRA_ENV"
    local create_venv=true
    local venv_options
    if $OFFLINE_MODE; then
        venv_options="--no-download --extra-search-dir=$KOMPIRA_EXTRA_WHL"
    fi
    if [ -f $KOMPIRA_ENV/bin/activate ]; then
        # $KOMPIRA_ENV/bin/pip が python 2.6 用かチェックする
        local pip_version=$(sudo -u $KOMPIRA_USER $KOMPIRA_ENV/bin/pip --version)
        if echo "$pip_version" | grep -q "python.*2.6"; then
            echo_debug "pip for python 2.6 detected: $pip_version"
        else
            create_venv=false
        fi
    fi

    if $create_venv; then
        local KOMPIRA_USERBASE=$(sudo -u $KOMPIRA_USER $PYTHON -c 'import site; print(site.USER_BASE)')
        USER_SUDO PYTHONUSERBASE=$KOMPIRA_USERBASE https_proxy=$TEMP_PROXY_URL $KOMPIRA_USERBASE/bin/virtualenv $venv_options --python=$PYTHON --no-wheel $KOMPIRA_ENV
        exit_if_failed "$?" "Failed to create virtualenv"
    fi
    # virtualenv の pip 用 proxy 設定
    echo -e "[global]\nproxy=$PROXY_URL" > $TMPDIR/kompira_pip.conf
    if ! diff -N -u $KOMPIRA_ENV/pip.conf $TMPDIR/kompira_pip.conf; then
        USER_SUDO cp -S .old -b $TMPDIR/kompira_pip.conf $KOMPIRA_ENV/pip.conf
    fi
}

activate_virtualenv()
{
    source $KOMPIRA_ENV/bin/activate
}

deactivate_virtualenv()
{
    deactivate
}

install_python27()
{
    echo_title "Install Python 2.7 package"
    if [ $SYSTEM == "AMZN" ]; then
        #
        # AMZN Linux では python27 はインストール済みなはず
        #
        yum_install python27 python27-devel
        PYTHON=$(which python27 2>/dev/null)
    else
        local result=$($PYTHON -V 2>&1 >/dev/null)
        case $result in
            $REQUIRES_PYTHON27)
                if $OFFLINE_MODE; then
                    yum_install python
                else
                    yum_install python python-devel
                fi
                if $EXTRA_MODE; then
                    PACKAGES_FOR_OFFLINE="$PACKAGES_FOR_OFFLINE python"
                fi
                ;;
            *)
                if ! $OFFLINE_MODE; then
                    if [ $SYSTEM == "CENT" ]; then
                        echo_info "Install SCL repositories"
                        yum_install "$SCL_CENT_REPOPKG"
                    elif [ $SYSTEM == "RHEL" ]; then
                        if $RHUI_MODE; then
                            YUMOPT_PYTHON27="--enablerepo=${SCL_RHUI_REPONAME}"
                            # verbose_run yum-config-manager --enable $SCL_RHUI_REPONAME > /dev/null
                        else
                            YUMOPT_PYTHON27="--enablerepo=${SCL_RHEL_REPONAME}"
                            # verbose_run yum-config-manager --enable $SCL_RHEL_REPONAME > /dev/null
                        fi
                    fi
                fi
                #
                # SCL では python27 で python27-devel も合わせてインストールされる
                #
                if $OFFLINE_MODE; then
                    YUM_OPTION=$YUMOPT_PYTHON27 yum_install python27
                else
                    YUM_OPTION=$YUMOPT_PYTHON27 yum_install python27 python27-python-devel
                fi
                if $EXTRA_MODE; then
                    PACKAGES_FOR_OFFLINE="$PACKAGES_FOR_OFFLINE python27"
                fi
                verbose_run echo "$SCL_PYTHON27_ROOT/usr/lib64/" > $TMPDIR/scl-python27-x86_64.conf
                diff_cp $TMPDIR/scl-python27-x86_64.conf /etc/ld.so.conf.d/scl-python27-x86_64.conf
                verbose_run ldconfig
                PYTHON="$SCL_PYTHON27_ROOT/usr/bin/python2.7"
                ;;
        esac
    fi
    if [ ! -x $PYTHON ]; then
        abort_setup "$PYTHON not found!"
    fi
}

install_requires_prepare()
{
    echo_title "Install RPM packages for preparation"
    YUM_OPTION=$YUMOPT_INSTALL yum_install $PACKAGES_FOR_INSTALL
}

install_requires_sendevt()
{
    echo_title "Install RPM packages for Kompira-sendevt"
    YUM_OPTION=$YUMOPT_SENDEVT yum_install $PACKAGES_FOR_SENDEVT
}

install_requires_jobmngrd()
{
    echo_title "Install RPM packages for Kompira-jobmngr"
    #
    # オフラインモードでないときは必要なリポジトリを有効にする
    #
    local packages_for_sendevt=$PACKAGES_FOR_SENDEVT
    local packages_for_jobmngr=$PACKAGES_FOR_JOBMNGR
    local yum_options
    if ! $OFFLINE_MODE; then
        if [ $SYSTEM == "RHEL" ]; then
            if $RHUI_MODE; then
                # yum_options="$yum_options --enablerepo=$RHUI_OPTION_REPONAME"
                verbose_run yum-config-manager --enable $RHUI_OPTION_REPONAME > /dev/null
            else
                # yum_options="$yum_options --enablerepo=$RHEL_OPTION_REPONAME"
                verbose_run yum-config-manager --enable $RHEL_OPTION_REPONAME > /dev/null
            fi
        fi
        packages_for_jobmngr="$packages_for_jobmngr $PACKAGES_FOR_JOBMNGR_ONLINE"
    fi
    YUM_OPTION=$yum_options yum_install $packages_for_sendevt $packages_for_jobmngr
}

install_requires_kompira()
{
    echo_title "Install RPM packages for Kompira"
    #
    # オフラインモードでないときは必要なリポジトリをインストール／有効化する
    #
    local packages_for_sendevt=$PACKAGES_FOR_SENDEVT
    local packages_for_jobmngr=$PACKAGES_FOR_JOBMNGR
    local packages_for_kompira=$PACKAGES_FOR_KOMPIRA
    local packages_for_kompira_rabbitmq=$PACKAGES_FOR_KOMPIRA_RABBITMQ
    local packages_for_kompira_postgresql=$PACKAGES_FOR_KOMPIRA_POSTGRESQL
    local yum_options
    if ! $OFFLINE_MODE; then
        if [ $SYSTEM != "AMZN" ]; then
            echo_info "Install Yum repositories (EPEL/PGSQL)"
            install_repo_from_rpm "$EPEL_REPOPKG" "${EPEL_REPOPKG}*" $EPEL_YUMREPO
            install_repo_from_rpm "$PG_REPOPKG" "${PG_REPOPKG0}*" $PG_YUMREPO
        fi
        #yum_options="$yum_options --enablerepo=${EPEL_REPONAME}"
        verbose_run yum-config-manager --enable $EPEL_REPONAME > /dev/null
        if [ $SYSTEM == "RHEL" ]; then
            if $RHUI_MODE; then
                # yum_options="$yum_options --enablerepo=$RHUI_OPTION_REPONAME"
                verbose_run yum-config-manager --enable $RHUI_OPTION_REPONAME > /dev/null
            else
                # yum_options="$yum_options --enablerepo=$RHEL_OPTION_REPONAME"
                verbose_run yum-config-manager --enable $RHEL_OPTION_REPONAME > /dev/null
            fi
        fi
        if [ $SYSTEM != "AMZN" ]; then
            # yum_options="$yum_options --enablerepo=$PG_REPONAME"
            verbose_run yum-config-manager --enable $PG_REPONAME > /dev/null
        fi
        packages_for_jobmngr="$packages_for_jobmngr $PACKAGES_FOR_JOBMNGR_ONLINE"
        packages_for_kompira="$packages_for_kompira $PACKAGES_FOR_KOMPIRA_ONLINE"
        packages_for_kompira_rabbitmq="$packages_for_kompira_rabbitmq $PACKAGES_FOR_KOMPIRA_RABBITMQ_ONLINE"
        packages_for_kompira_postgresql="$packages_for_kompira_postgresql $PACKAGES_FOR_KOMPIRA_POSTGRESQL_ONLINE"
    fi

    YUM_OPTION=$yum_options yum_install $packages_for_sendevt $packages_for_jobmngr \
              $packages_for_kompira $packages_for_kompira_rabbitmq $packages_for_kompira_postgresql
}

start_kompirad()
{
    _service_start kompirad
}

start_kompira_jobmngrd()
{
    _service_start kompira_jobmngrd
}

restart_kompirad()
{
    _service_restart kompirad
}

restart_kompira_jobmngrd()
{
    _service_restart kompira_jobmngrd
}

stop_kompirad()
{
    _service_stop kompirad
}

stop_kompira_jobmngrd()
{
    _service_stop kompira_jobmngrd
}

stop_kompira()
{
    if $CLUSTER_RUNNING; then return 0; fi
    echo_title "Stop Kompira daemon and Kompira-jobmngr daemon"
    stop_kompirad
    stop_kompira_jobmngrd
}

uninstall_kompira_package()
{
    local -x PIP=$(which pip 2> /dev/null)
    if [ -z "$PIP" ]; then
        return 0
    fi
    echo_title "Uninstall old Kompira packages [pip=$PIP]"
    pip_uninstall_if_installed $KOMPIRA_PKGNAME_DEFAULT $KOMPIRA_PKGNAME_JOBMNGR $KOMPIRA_PKGNAME_SENDEVT
}

uninstall_kompira()
{
    #
    # localenv 環境以前の kompira をアンインストール
    #
    uninstall_kompira_package

    #
    # localenv 環境対応の kompira をアンインストール
    #
    local -x PATH=$KOMPIRA_BIN:$PATH
    uninstall_kompira_package
}

install_whl_package()
{
    local pkg_file=$1
    local pkg_name=$(basename $pkg_file)
    local pip_options
    if $OFFLINE_MODE; then
        pip_options="--no-index --find-links=$KOMPIRA_EXTRA_WHL"
    fi
    if [ -n "$TEMP_PROXY_URL" ]; then
        pip_options="$pip_options --proxy=$TEMP_PROXY_URL"
    fi
    #
    # sudo -u kompira pip が展開した whl ファイルにアクセスできないので一旦 /tmp にコピーする
    # pypi から取得できないパッケージは wheelhouse/ に同梱する
    # pip install でインストールするパッケージのビルト時に必要になる場合があるので PATH 環境変数を渡す
    #
    local user_tmpdir=$(sudo -u $KOMPIRA_USER mktemp -d)
    verbose_run cp $pkg_file $user_tmpdir/$pkg_name
    verbose_run cp wheelhouse/*.whl $user_tmpdir/
    (cd / && USER_SUDO PATH=$PATH $KOMPIRA_BIN/pip install $pip_options --find-links=$user_tmpdir $user_tmpdir/$pkg_name)
    local rc=$?
    verbose_run rm -rf $user_tmpdir
    return $rc
}

install_package_sendevt()
{
    echo_title "Install Kompira-sendevt package"
    install_whl_package $KOMPIRA_WHLFILE_SENDEVT
    exit_if_failed "$?" "Failed to install Kompira-sendevt package"
}

install_package_jobmngr()
{
    echo_title "Install Kompira-jobmngr package"
    install_whl_package $KOMPIRA_WHLFILE_JOBMNGR
    exit_if_failed "$?" "Failed to install Kompira-jobmngr package"
}

install_package_default()
{
    echo_title "Install Kompira package"
    install_whl_package $KOMPIRA_WHLFILE_DEFAULT
    exit_if_failed "$?" "Failed to install Kompira package"
    # setup_utils.sh の VERSION/BRANCH を更新する
    # TODO: 本来は whl パッケージ作成時に更新したい
    verbose_run sed -i -e "/^BRANCH=/c BRANCH=$BRANCH" -e "/^VERSION=/c VERSION=$VERSION" -e "/^SHORT_VERSION=/c SHORT_VERSION=$SHORT_VERSION" $KOMPIRA_BIN/setup_utils.sh
}

backup_kompira_database()
{
    if $CLUSTER_RUNNING; then return 0; fi
    if ! $BACKUP_MODE; then return 0; fi
    if [ ! -x $KOMPIRA_BIN/manage.py ]; then
        return 0
    fi
    if [ -z "$OLD_PG_VERSION" ]; then
        return 0
    fi
    echo_title "Backup kompira database"
    if ! pgsql_has_kompira_table; then
        echo_info "Table is not exist. (SKIP)"
        return 0
    fi
    if [ -f $KOMPIRA_BACKUP ]; then
        echo_info "File exists already: $KOMPIRA_BACKUP"
        local index=1
        while [ -f "$KOMPIRA_BACKUP.$index" ]; do
            let index=$index+1
        done
        verbose_run sudo -u $KOMPIRA_USER mv $KOMPIRA_BACKUP "$KOMPIRA_BACKUP.$index"
    fi
    # loaddata 時にエラーになる contenttypes, auth.permission は除外する
    local DUMPDATA_EXCLUDES="-e contenttypes -e auth.permission -e sessions"
    # kompira1.4/django1.6 系の場合は south も除外する
    if $KOMPIRA_BIN/manage.py --help | grep -qF "[south]"; then
        DUMPDATA_EXCLUDES+=" -e south"
    fi
    # kompira 1.5.1 未満の場合は process を適切に dumpdata できないので除外する
    if [ $(ver2num $OLD_KOMPIRA_VERSION) -lt $(ver2num "1.5.1") ]; then
        DUMPDATA_EXCLUDES+=" -e process"
    elif ! $BACKUP_PROCESS; then
        DUMPDATA_EXCLUDES+=" -e process"
    fi
    verbose_run $KOMPIRA_BIN/manage.py dumpdata --traceback $DUMPDATA_EXCLUDES | su $KOMPIRA_USER -c "gzip > $KOMPIRA_BACKUP"
    exit_if_failed "$?" "Failed to backup database"
    # INITDATA_MODE のときだけ自動リストアする
    if $INITDATA_MODE; then
        USER_INSTALL -g $KOMPIRA_GROUP -m 775 -d $KOMPIRA_VAR_DIR/startup
        sudo -u $KOMPIRA_USER touch $KOMPIRA_VAR_DIR/startup/do_restore
    fi
}

restore_kompira_database()
{
    echo_info "Restore kompira database"
    if [ ! -f $KOMPIRA_BACKUP ]; then
        echo_warn "File not found: $KOMPIRA_BACKUP"
        return 0
    fi
    verbose_run $KOMPIRA_BIN/manage.py loaddata --ignorenonexistent $KOMPIRA_BACKUP
    exit_if_failed "$?" "Failed to restore database"
}

setup_pgsql()
{
    if $CLUSTER_RUNNING; then return 0; fi
    echo_title "Configure PostgreSQL settings"
    #
    # 動作中の古い postgresql を停止/無効化する
    #
    if [[ -n "$OLD_PG_SERVICE" && "$OLD_PG_SERVICE" != "$PG_SERVICE" ]]; then
        echo_info "Disable postgresql ${OLD_PG_VERSION}"
        _service_stop $OLD_PG_SERVICE
        service_disable $OLD_PG_SERVICE
    fi
    #
    # [AMZN] PGDATA と PGLOG のデフォルト値を CentOS/RHEL に合わせる
    #
    if [ $SYSTEM == "AMZN" ]; then
        cat > $TMPDIR/$PG_SERVICE <<EOF
# Set defaults for configuration variables
PGDATA=$PG_DATADIR
PGLOG=$PG_BASEDIR/pgstartup.log
EOF
        diff_cp $TMPDIR/$PG_SERVICE /etc/sysconfig/pgsql/$PG_SERVICE
    fi
    #
    # PGDATA/PG_VERSION が無い場合 initdb を実施する
    # パラメータでロケールを指定する
    #
    if [ ! -f $PG_DATADIR/PG_VERSION ]; then
        if $SYSTEMD; then
            PGSETUP_INITDB_OPTIONS="--locale=$LOCALE_LANG" verbose_run $(pgsql_cmddir)/postgresql${PG_VERSION/./}-setup initdb
        else
            verbose_run service $PG_SERVICE initdb $LOCALE_LANG
        fi
    fi
    #
    # postgresql サービスを開始する
    # local からのアクセスを認証なしで許可する
    # 行頭が local で始まる行中の文字列 peer を trust に置換する
    #
    service_start $PG_SERVICE
    verbose_run sed -e "/^local/{s/peer/trust/}" $PG_DATADIR/pg_hba.conf > $TMPDIR/pg_hba.conf
    diff_cp $TMPDIR/pg_hba.conf $PG_DATADIR/pg_hba.conf
    service_restart $PG_SERVICE

    #
    # INITDATA モードでは database を完全初期化する
    #
    if $INITDATA_MODE; then
        echo_info "Delete Kompira Database: $KOMPIRA_PG_DATABASE"
        verbose_run dropdb -U $PGSQL_USER --if-exists -e $KOMPIRA_PG_DATABASE
        echo_info "Delete Kompira DB-user: $KOMPIRA_PG_USER"
        verbose_run dropuser -U $PGSQL_USER --if-exists -e $KOMPIRA_PG_USER
    fi
    #
    # データベースユーザ $KOMPIRA_PG_USER を作成する
    #
    if ! psql -U $KOMPIRA_PG_USER -d $PGSQL_DB -c \\q; then
        echo_info "Create Kompira DB-user: $KOMPIRA_PG_USER"
        verbose_run createuser -U $PGSQL_USER --createdb -e $KOMPIRA_PG_USER
        exit_if_failed "$?" "Failed to create PostgreSQL user '$KOMPIRA_PG_USER'"
    fi
    #
    # データベース $KOMPIRA_PG_DATABASE を作成する
    #
    if ! psql -U $PGSQL_USER -d $KOMPIRA_PG_DATABASE -c \\q; then
        echo_info "Create Kompira Database: $KOMPIRA_PG_DATABASE"
        verbose_run createdb -U $KOMPIRA_PG_USER --encoding=utf8 -e $KOMPIRA_PG_DATABASE
        exit_if_failed "$?" "Failed to create PostgreSQL database '$KOMPIRA_PG_DATABASE'"
    fi

    if ! $CLUSTER_CONFIGURED; then
        #
        # シングル構成のときだけ pgsql を有効化／起動したままにする
        #
        service_enable $PG_SERVICE
    else
        #
        # 冗長構成のときは pgsql を停止させておく
        #
        service_stop $PG_SERVICE
    fi
}

setup_apache()
{
    if $CLUSTER_RUNNING; then return 0; fi
    echo_title "Configure apache settings."
    verbose_run $KOMPIRA_BIN/mod_wsgi-express install-module > $TMPDIR/mod_wsgi.conf
    exit_if_failed "$?" "Failed to install mod_wsgi"
    diff_cp $TMPDIR/mod_wsgi.conf /etc/httpd/conf.d/mod_wsgi.conf

    local httpd_conf=/etc/httpd/conf.d/kompira.conf
    local site_packages=$($KOMPIRA_BIN/python -c "import django; print django.__path__[0]" | sed -re 's|(.*/site-packages).*|\1|')
    local comment_if_http
    local template_prefix
    [ $HTTPS_MODE != true ] && comment_if_http="# " || comment_if_http=""
    local httpd_version=$(rpm -q httpd)
    case "$httpd_version" in
        httpd-2.2.*) template_prefix="httpd22" ;;
        httpd-2.4.*) template_prefix="httpd24" ;;
        *) abort_setup "Unexpected httpd version: $httpd_version" ;;
    esac

    echo_info "Append kompira configure ($template_prefix)."
    expand_template $THIS_DIR/config/httpd/$template_prefix-kompira.conf > $TMPDIR/kompira.conf
    diff_cp $TMPDIR/kompira.conf $httpd_conf

    echo_info "Change LANG setting to '$LOCALE_LANG'."
    verbose_run sed -re "s/^#?\s*(HTTPD_LANG|LANG)=.*/\1=$LOCALE_LANG/" /etc/sysconfig/httpd > $TMPDIR/httpd
    echo_info "Change proxy settings to '$PROXY_URL'."
    local env_export=""
    if ! $SYSTEMD; then
        env_export="export "
    fi
    if grep -E -i -q "^${env_export}http?_proxy=" $TMPDIR/httpd; then
        verbose_run \
            sed -i -r \
            -e "s|^(${env_export}https?_proxy)=.*|\1=\"$PROXY_URL\"|i" \
            -e "s|^(${env_export}no_proxy)=.*|\1=\"$NO_PROXY\"|" \
            $TMPDIR/httpd
    else
        cat >> $TMPDIR/httpd <<EOF

# proxy settings
${env_export}http_proxy="$PROXY_URL"
${env_export}https_proxy="$PROXY_URL"
${env_export}HTTP_PROXY="$PROXY_URL"
${env_export}HTTPS_PROXY="$PROXY_URL"
${env_export}no_proxy="$NO_PROXY"
EOF
    fi
    diff_cp $TMPDIR/httpd /etc/sysconfig/httpd

    echo_info "Append user apache to group $KOMPIRA_GROUP."
    verbose_run usermod -G $KOMPIRA_GROUP apache

    echo_info "Disable default welcome page."
    verbose_run sed -re 's/^([^#])/# \1/' /etc/httpd/conf.d/welcome.conf > $TMPDIR/welcome.conf
    diff_cp $TMPDIR/welcome.conf /etc/httpd/conf.d/welcome.conf

    if ! $CLUSTER_CONFIGURED; then
        #
        # シングル構成のときだけ apache を有効化／起動する
        #
        echo_info "Enable and restart apache service"
        service_enable httpd
        service_restart httpd
    fi
}

setup_rabbitmq()
{
    if $CLUSTER_RUNNING; then return 0; fi
    echo_title "Configure rabbitmq settings"

    if ! $CLUSTER_CONFIGURED; then
        #
        # シングル構成のときだけ rabbitmq-server を有効化／起動する
        #
        service_enable rabbitmq-server
        service_restart rabbitmq-server
    fi
}

setup_firewall()
{
    if $CLUSTER_RUNNING; then return 0; fi
    if is_activce_service firewalld; then
        setup_firewalld $FIREWALLD_ACCEPT_PORTS
    elif is_activce_service iptables; then
        setup_iptables $IPTABLES_ACCEPT_PORTS
    fi
}

setup_selinux()
{
    if $CLUSTER_RUNNING; then return 0; fi
    echo_title "Configure SELinux to $SELINUX_MODE."
    # setsebool -P httpd_can_network_connect on
    verbose_run setenforce $SELINUX_MODE
    verbose_run sed -e "/^SELINUX=/c SELINUX=$SELINUX_MODE" /etc/selinux/config > $TMPDIR/selinux
    diff_cp $TMPDIR/selinux /etc/selinux/config
}

setup_kompira_common()
{
    echo_title "Setup kompira common files."
    #
    # kompira グループに追加した apache アカウントで動作する httpd
    # からもログを書きこむので、パーミッションに 775 を設定する
    # 既存のログで所有者が root なものは kompira に変更する
    #
    echo_info "Create log directory: $KOMPIRA_LOG_DIR"
    USER_INSTALL -g $KOMPIRA_GROUP -m 775 -d $KOMPIRA_LOG_DIR
    verbose_run find $KOMPIRA_LOG_DIR -type f -user root -exec chown $KOMPIRA_USER:$KOMPIRA_GROUP {} \;

    echo_info "Create Kompira/Kompira-sendevt configuration file."
    expand_template $THIS_DIR/config/kompira.conf > $TMPDIR/kompira.conf
    if ! diff -N -u $KOMPIRA_HOME/kompira.conf $TMPDIR/kompira.conf; then
        USER_INSTALL -S .old -b $TMPDIR/kompira.conf -m 644 $KOMPIRA_HOME/kompira.conf
    fi
}

setup_kompira_jobmngr()
{
    echo_title "Setup kompira-jobmngrd."
    if $SYSTEMD; then
        echo_info "[systemd] Create kompira-jobmngrd configuration file: /etc/sysconfig/kompira_jobmngrd"
        expand_template $THIS_DIR/config/systemd/kompira_jobmngrd > $TMPDIR/kompira_jobmngrd
        diff_cp $TMPDIR/kompira_jobmngrd /etc/sysconfig/kompira_jobmngrd

        echo_info "[systemd] Enable kompira-jobmngrd service."
        expand_template $THIS_DIR/config/systemd/kompira_jobmngrd.service > $TMPDIR/kompira_jobmngrd.service
        verbose_run $INSTALL -m 664 $TMPDIR/kompira_jobmngrd.service /usr/lib/systemd/system/kompira_jobmngrd.service
        service_register kompira_jobmngrd
    else
        echo_info "[upstart] Create kompira-jobmngrd configuration file: /etc/init/kompira_jobmngrd.conf"
        expand_template $THIS_DIR/config/upstart/kompira_jobmngrd.conf > $TMPDIR/kompira_jobmngrd.conf
        diff_cp $TMPDIR/kompira_jobmngrd.conf /etc/init/kompira_jobmngrd.conf
    fi

    if ! $CLUSTER_CONFIGURED; then
        #
        # シングル構成のときだけ kompira_jobmngrd を有効化／起動する
        #
        echo_info "Start kompira-jobmngrd service."
        service_enable kompira_jobmngrd
        restart_kompira_jobmngrd
    fi
}

setup_kompira_datafiles()
{
    echo_title "Setup kompira data files."
    #
    # kompira グループに追加した apache アカウントで動作する httpd
    # からもアクセスするので、パーミッションに 775 を設定する
    #
    echo_info "Collect static files in kompira data directory."
    if $INITFILE_MODE; then
        verbose_run rm -rf $KOMPIRA_VAR_DIR/upload $KOMPIRA_VAR_DIR/repository
    fi
    USER_INSTALL -g $KOMPIRA_GROUP -m 775 -d $KOMPIRA_VAR_DIR{,/exported,/upload,/repository,/startup}
    #
    # exported/*.dat ファイルコピー
    #
    USER_INSTALL -g $KOMPIRA_GROUP -m 644 $THIS_DIR/exported/{types-*.dat,startup-*.dat} $KOMPIRA_VAR_DIR/exported
    #
    # collectstatic はシングル、マスター、スレーブいずれでも、この段階で実施する
    #
    USER_SUDO $KOMPIRA_BIN/manage.py collectstatic --noinput
    #
    # pdf 形式のマニュアルをコピーする
    #
    local pdf locale
    for pdf in Kompira-*.pdf; do
        local locale=${pdf/.*}
        locale=${locale#*-}
        if [ -n "$locale" ]; then
            if [ -d $KOMPIRA_VAR_DIR/html/docs/$locale ]; then
                $INSTALL -m 644 $pdf $KOMPIRA_VAR_DIR/html/docs/$locale/kompira.pdf
            else
                echo_warn "directory not found: $KOMPIRA_VAR_DIR/html/docs/$locale"
            fi
        fi
    done

    #
    # 次回 kompirad 起動時にデータベースの初期化を実施させるため、
    # $KOMPIRA_VAR_DIR/startup/ にファイルを作成しておく
    #
    sudo -u $KOMPIRA_USER touch $KOMPIRA_VAR_DIR/startup/{do_migrate,do_import_data,do_compile_jobflow}
}

setup_kompira_settings()
{
    echo_title "Setup kompira settings."
    local -x PIP=$(which pip 2> /dev/null)
    local kompira_location=$(get_kompira_location)
    local kompira_settings="$kompira_location/kompira/settings.py"
    if [ ! -f $kompira_settings ]; then
        echo_warn "File not found: $kompira_settings"
        return 0
    fi
    local language_code='en-us'
    case "$LOCALE_LANG" in
        "") ;;
        ja*) language_code="ja" ;;
        *) language_code=$(echo "${LOCALE_LANG/.*/}" | tr A-Z_ a-z-) ;;
    esac
    echo_debug "language_code=$language_code (LOCALE_LANG=$LOCALE_LANG)"
    sed -re "s/^(LANGUAGE_CODE)\s+=.*/\1 = \'$language_code\'/" \
        $kompira_settings > $TMPDIR/kompira_settings.py
    diff_cp $TMPDIR/kompira_settings.py $kompira_settings
}

setup_kompira_daemon()
{
    echo_title "Setup kompira-daemon."
    if $SYSTEMD; then
        echo_info "[systemd] Create kompira-daemon configuration file: /etc/sysconfig/kompirad"
        expand_template $THIS_DIR/config/systemd/kompirad > $TMPDIR/kompirad
        diff_cp $TMPDIR/kompirad /etc/sysconfig/kompirad

        echo_info "[systemd] Enable kompira-daemon service."
        expand_template $THIS_DIR/config/systemd/kompirad.service > $TMPDIR/kompirad.service
        verbose_run $INSTALL -m 664 $TMPDIR/kompirad.service /usr/lib/systemd/system/kompirad.service
        service_register kompirad
    else
        echo_info "[upstart] Create kompira-daemon configuration file: /etc/init/kompirad.conf"
        expand_template $THIS_DIR/config/upstart/kompirad.conf > $TMPDIR/kompirad.conf
        diff_cp $TMPDIR/kompirad.conf /etc/init/kompirad.conf

        echo_info "[sysvinit] Enable kompira-daemon service."
        verbose_run $INSTALL -m 755 $THIS_DIR/config/upstart/kompirad /etc/init.d/
        service_register kompirad
    fi

    if ! $CLUSTER_CONFIGURED; then
        #
        # シングル構成のときだけ kompirad を有効化／起動する
        #
        echo_info "Start kompira-daemon service."
        service_enable kompirad
        restart_kompirad
    fi
}

setup_kompira_default()
{
    setup_kompira_datafiles
    setup_kompira_settings
    setup_kompira_daemon
}

update_cluster()
{
    if ! $CLUSTER_RUNNING; then return 0; fi
    echo
    echo "*** Update clustering settings."
    echo
    # TODO: 以下は暫定対処であり、設定管理の方式について見直す
    # monitor 動作の設定更新
    if ! pcs resource show res_rabbitmq | grep -q "res_rabbitmq-monitor-interval-7s"; then
        verbose_run pcs resource op remove res_rabbitmq monitor
        verbose_run pcs resource op add res_rabbitmq monitor interval="7s" timeout="20s"
        verbose_run pcs resource op add res_rabbitmq monitor interval="5s" timeout="20s" role="Master"
    fi
    if ! pcs resource show res_pgsql | grep -q "res_pgsql-monitor-interval-7s"; then
        verbose_run pcs resource op remove res_pgsql monitor
        verbose_run pcs resource op add res_pgsql monitor interval="7s" timeout="20s"
        verbose_run pcs resource op add res_pgsql monitor interval="5s" timeout="20s" role="Master"
    fi
    # 制約条件の設定更新
    if ! pcs constraint show | grep -q "set webserver set ms_rabbitmq ms_pgsql"; then
        pcs_remove_constraints
        pcs_setup_constraints
    fi
}

test_kompira()
{
    if $CLUSTER_CONFIGURED; then return 0; fi
    echo_title "Test access to kompira."
    verbose_run curl -ksf $KOMPIRA_TEST_URL > $TMPDIR/test-kompira.out
    if [ $? != 0 ]; then
        cat $TMPDIR/test-kompira.out
        abort_setup "Access Failed!"
    fi
    echo_info "Access succeeded: $(grep "brand-version" $TMPDIR/test-kompira.out)"
}

install_kompira_sendevt()
{
    install_requires_sendevt
    if $REQUIRES_ONLY; then return 0; fi
    install_package_sendevt
    setup_kompira_common
}

install_kompira_jobmngr()
{
    install_requires_jobmngrd
    if $REQUIRES_ONLY; then return 0; fi
    install_package_jobmngr
    setup_kompira_common
    setup_kompira_jobmngr
}

install_kompira_default()
{
    install_requires_kompira
    if $REQUIRES_ONLY; then return 0; fi
    install_package_default
    setup_pgsql
    setup_apache
    setup_rabbitmq
    setup_firewall
    setup_selinux
    setup_kompira_common
    setup_kompira_jobmngr
    setup_kompira_default
    update_cluster
    test_kompira
}

create_extra()
{
    local requirements
    local repotrack_options
    if [ -f scripts/requires.txt ]; then
        requirements=scripts/requires.txt
    elif [ -f requirements/kompira.txt ]; then
        requirements=requirements/kompira.txt
    else
        abort_setup "requirements not found!"
    fi
    eval $(grep -E "^(RHEL_HA_REPONAME|PACKAGES_FOR_HA_EL)" setup_cluster.sh)
    if [ $RHEL_VERSION == 6 ]; then
        PACKAGES_FOR_HA=$PACKAGES_FOR_HA_EL6
    elif [ $RHEL_VERSION == 7 ]; then
        PACKAGES_FOR_HA=$PACKAGES_FOR_HA_EL7
    fi
    if [ $SYSTEM == "RHEL" ]; then
        if $RHUI_MODE; then
            install_vault_repo
            verbose_run yum-config-manager --enable $VAULT_CENTOS_REPONAME > /dev/null
            verbose_run yum-config-manager --enable $SCL_RHUI_REPONAME > /dev/null
        else
            verbose_run yum-config-manager --enable $SCL_RHEL_REPONAME > /dev/null
            verbose_run yum-config-manager --enable $RHEL_HA_REPONAME > /dev/null
        fi
    fi

    echo_title "Create extra package"
    verbose_run mkdir -p $KOMPIRA_EXTRA_DIR/{/etc/yum.repos.d,$KOMPIRA_EXTRA_PIP,$KOMPIRA_EXTRA_RPM,$KOMPIRA_EXTRA_WHL}

    echo_info "Install rpm packages for create extra package"
    yum_install $PACKAGES_FOR_EXTRA
    install_repo_from_rpm "$EPEL_REPOPKG" "${EPEL_REPOPKG}*" $EPEL_YUMREPO
    install_repo_from_rpm "$PG_REPOPKG" "${PG_REPOPKG0}*" $PG_YUMREPO

    echo_info "Download rpm packages"
    verbose_run repotrack -p $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_RPM \
                $PACKAGES_FOR_INSTALL $PACKAGES_FOR_OFFLINE \
                $PACKAGES_FOR_SENDEVT $PACKAGES_FOR_JOBMNGR $PACKAGES_FOR_KOMPIRA \
                $PACKAGES_FOR_KOMPIRA_RABBITMQ $PACKAGES_FOR_KOMPIRA_POSTGRESQL \
                $PACKAGES_FOR_HA
    exit_if_failed "$?" "Failed to download rpm packages"
    createrepo $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_RPM
    cat > $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_REPO <<EOF
[$KOMPIRA_EXTRA_NAME]
name=$KOMPIRA_EXTRA_NAME
baseurl=file://$KOMPIRA_EXTRA_RPM
gpgcheck=0
enabled=0
EOF

    echo_info "Download get-pip.py and install pip to userbase"
    verbose_run curl -ksL -o $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_PIP/get-pip.py $GETPIP_PY
    verbose_run $PYTHON $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_PIP/get-pip.py -I -U --user

    echo_info "Download wheel packages"
    verbose_run ~/.local/bin/pip wheel -w $KOMPIRA_EXTRA_DIR/$KOMPIRA_EXTRA_WHL --find-links=./wheelhouse -r $requirements pip setuptools wheel virtualenv
    exit_if_failed "$?" "Failed to create wheelhouse"

    echo_info "Packaging extra package"
    verbose_run tar zcvf $KOMPIRA_EXTRA_DIR.tar.gz -C $KOMPIRA_EXTRA_DIR etc opt
}

install_kompira()
{
    if $SENDEVT_MODE; then
        install_kompira_sendevt
    elif $JOBMNGR_MODE; then
        install_kompira_jobmngr
    else
        install_kompira_default
    fi
}

usage_exit()
{
    cat <<EOF
Usage: $(basename $0) [OPTIONS] <kompira-server>

  --force                       Force install.
  --https                       Configure apache to force secure access (default).
  --no-https                    Configure apache to enable unsecure access.
  --initdata                    Initialize kompira database.
  --initfile                    Initialize $KOMPIRA_VAR_DIR/{upload,repository}.
  --no-backup                   Skip database backup.
  --no-locale                   Skip configure system locale.
  --locale-lang <lang>          Configure system locale laguage.
  --locale-timezone <zonename>  Configure system locale timezone.
  --proxy <proxy>               Specify a proxy in the form
                                [user:passwd@]proxy.server:port.
  --noproxy <hosts>             Comma-separated list of hosts which do not use proxy.
  --jobmngr                     Install Kompira job manager package.
  --sendevt                     Install Kompira send event package.
  --help                        Show this message.

EOF
    exit 1
}

parse_options()
{
    SPECIFY_SERVER=false
    OPTIONS=`getopt -q -o '' -l https,no-https,force,initdata,initfile,no-backup,backup-process,update,migrate-fake,rhui,extra,offline,no-locale,locale-lang:,locale-timezone:,temp-proxy:,temp-noproxy:,proxy:,noproxy:,jobmngr,sendevt,user,requires-only,dry-run,help -- "$@"`
    [ $? != 0 ] && usage_exit
    eval set -- "$OPTIONS"
    while true
    do
        arg="$1"
        case "$arg" in
            --https) HTTPS_MODE=true ;;
            --no-https) HTTPS_MODE=false ;;
            --force) FORCE_MODE=true ;;
            --initdata) INITDATA_MODE=true ;;
            --initfile) INITFILE_MODE=true ;;
            --no-backup) BACKUP_MODE=false ;;
            --backup-process) BACKUP_PROCESS=true ;;
            --migrate-fake) MIGRATE_FAKE=true ;;
            --rhui) RHUI_MODE=true ;;
            --extra) EXTRA_MODE=true
                     SETUP_TYPE="Create kompira-extra package" ;;
            --offline) OFFLINE_MODE=true
                       SETUP_TYPE="$SETUP_TYPE (offline-mode)" ;;
            --no-locale) CONFIG_LOCALE_LANG=false; CONFIG_LOCALE_TIMEZONE=false ;;
            --locale-lang) CONFIG_LOCALE_LANG=true; LOCALE_LANG=$2; shift ;;
            --locale-timezone) CONFIG_LOCALE_TIMEZONE=true; LOCALE_TIMEZONE=$2; shift ;;
            --temp-proxy) TEMP_PROXY_URL=$(normalize_proxy_url $2); shift;;
            --temp-noproxy) TEMP_NO_PROXY=$2; shift ;;
            --proxy) PROXY_URL=$(normalize_proxy_url $2); TEMP_PROXY_URL=$PROXY_URL; shift ;;
            --noproxy) NO_PROXY=$2; TEMP_NO_PROXY=$NO_PROXY; shift ;;
            --jobmngr) JOBMNGR_MODE=true; SPECIFY_SERVER=true ;;
            --sendevt) SENDEVT_MODE=true; SPECIFY_SERVER=true ;;
            --user) USER_MODE=true ;;
            --requires-only) REQUIRES_ONLY=true ;;
            --dry-run) DRY_RUN=true ;;
            --update)
                echo "option $arg is deprecated."
                ;;
            --) shift; break ;; # 引数は無視
            *) usage_exit ;;
        esac
        shift
    done

    if $JOBMNGR_MODE && $SENDEVT_MODE; then
        echo '--jobmngr and --sendevt options are exclusive.'
        echo
        exit 1
    fi

    if $SPECIFY_SERVER; then
        if [ -z "$1" ]; then
            echo 'must specify <kompira-server>'
            echo
            usage_exit
        fi
        KOMPIRA_SERVER="$1"
    else
        if [ -n "$1" ]; then
            echo "unknown argument \"$1\""
            echo
            usage_exit
        fi
    fi
}

################################################################

set -o pipefail
parse_options "$@"
exec > >(tee $INSTALL_LOG) 2>&1
create_tmpdir
start_setup

#
# 環境チェック
#
check_root
check_environ
check_cluster
check_python
check_update
if $DRY_RUN; then
    exit_setup
fi
install_requires_prepare
yum_clean

#
# extra パッケージの作成
#
if $EXTRA_MODE; then
    create_extra
    exit_setup
fi

#
# システム設定
#
configure_locale_lang
configure_locale_timezone

#
# kompira アカウント/ホームディレクトリを作成する
#
create_user
create_home

#
# データベースのバックアップ
#
backup_kompira_database

#
# 動作中の kompira を停止し、アンインストールする
#
stop_kompira
uninstall_kompira

#
# root のユーザベース環境に pip/virtualenv をインストールする
# virtualenv で仮想環境を作成する
#
create_userbase
create_virtualenv

#
# virtualenv 環境をアクティブにしてインストールを実施する
#
activate_virtualenv
install_kompira
deactivate_virtualenv

exit_setup
